# built-in imports
import hashlib
import logging
from typing import Generator, List, Optional, Union
from pathlib import Path
import time
from hashlib import md5
from enum import Enum
from collections import Counter
import datetime
import random
import queue

# Third-party imports
import psutil
from libpastis import BrokerAgent, FuzzingEngineDescriptor, EngineConfiguration, BinaryPackage, SASTReport, ClientAgent
from libpastis.types import SeedType, FuzzingEngineInfo, LogLevel, Arch, State, SeedInjectLoc, CheckMode, CoverageMode, \
                            ExecMode, AlertData, PathLike, Platform, FuzzMode
from libpastis.utils import get_local_architecture
import lief
from tritondse import QuokkaProgram

# Local imports
from pastisbroker.client import PastisClient
from pastisbroker.stat_manager import StatManager
from pastisbroker.workspace import Workspace, WorkspaceStatus
from pastisbroker.utils import load_engine_descriptor, Bcolors, COLORS
from pastisbroker.coverage import CoverageManager, ClientInput


lief.logging.disable()


class BrokingMode(Enum):
    FULL = 1              # Transmit all seeds to all peers
    NO_TRANSMIT = 2       # Does not transmit seed to peers (for comparing perfs of tools against each other)




class PastisBroker(BrokerAgent):

    def __init__(self, workspace: PathLike,
                 binaries_dir: PathLike,
                 broker_mode: BrokingMode,
                 check_mode: CheckMode = CheckMode.CHECK_ALL,
                 inject_loc: SeedInjectLoc = SeedInjectLoc.STDIN,
                 sast_report: PathLike = None,
                 p_argv: List[str] = None,
                 memory_threshold: int = 85,
                 start_quorum: int = 0,
                 filter_inputs: bool = False,
                 stream: bool = False,
                 replay_threads: int = 4):
        super(PastisBroker, self).__init__()

        # Initialize workspace
        self.workspace = Workspace(Path(workspace))
        params = {"binaries_dir": str(Path(binaries_dir).absolute()),
                  "broker_mode": broker_mode.name,
                  "check_mode": check_mode.name,
                  "inject_loc": inject_loc.name,
                  "argvs": p_argv}
        self.workspace.initialize_runtime(binaries_dir, params)

        self._configure_logging()

        # Register all agent callbacks
        self._register_all()

        # Init internal state
        self.broker_mode = broker_mode
        self.ck_mode = check_mode
        self.inject = inject_loc
        self.argv = [] if p_argv is None else p_argv
        self.engines_args = {}
        self.engines = {}  # name->FuzzingEngineDescriptor

        # for slicing mode (otherwise not used)
        self._slicing_ongoing = {}  # Program -> {Addr -> [cli]}

        # Initialize availables binaries
        self.programs = {}  # Tuple[(Platform, Arch)] -> List[BinaryPackage]
        self._find_binaries(binaries_dir)

        # Klocwork informations
        self.sast_report = None
        if sast_report:
            self.initialize_sast_report(sast_report)

        # Client infos
        self.clients = {}   # bytes -> Client
        self._cur_id = 0

        # Runtime infos
        self._running = False
        self._seed_pool = {}  # Seed bytes -> SeedType
        self._init_seed_pool = {}  # Used for NO_TRANSMIT mode
        self._start_time = None
        self._stop = False

        # Load the workspace seeds
        self._load_workspace()

        # Create the stat manager
        self.statmanager = StatManager(self.workspace)

        # Watchdog to monitor RAM usage
        self.watchdog = None
        self._threshold = memory_threshold # percent

        # startup quorum
        self._startup_quorum = start_quorum
        self._current_quorum = 0

        # Proxy feature
        self._proxy = None
        self._proxy_cli = None
        self._proxy_start_signal = False
        self._proxy_seed_queue = queue.Queue()

        # Coverage + filtering feature
        self._coverage_manager = None
        self.filter_inputs: bool = filter_inputs
        if filter_inputs or stream:
            if (path := self.find_vanilla_binary()) is not None:  # Find an executable suitable for coverage
                logging.info(f"Coverage binary: {path}")
                stream_file = self.workspace.coverage_history if stream else ""
                self._coverage_manager = CoverageManager(replay_threads, filter_inputs, path, self.argv, self.inject, stream_file)
            else:
                logging.warning("filtering or stream enabled but cannot find vanilla binary")


    def find_vanilla_binary(self) -> Optional[str]:
        """
        Find a binary without instrumentation to be used for coverage
        computation. It also has to match local architecture.
        :return: Path to the progam
        """
        local_arch = get_local_architecture()
        for pkg in self.programs.get((Platform.LINUX, local_arch)):
            path = str(pkg.executable_path.absolute())
            p = lief.parse(str(path))
            ok = True
            for f in p.functions:
                if "hfuzz_" in f.name or "__afl_" in f.name or "__gcov_" in f.name or "__asan_" in f.name:
                    ok = False
                    break
            if ok:
                return path
        return None


    def load_engine_addon(self, py_module: str) -> bool:
        desc = load_engine_descriptor(py_module)
        if desc is not None:
            self.engines[desc.NAME] = desc
            self.engines_args[desc.NAME] = []
            return True
        else:
            return False

    def initialize_sast_report(self, report: PathLike):
        self.sast_report = SASTReport.from_file(report)
        self.workspace.add_sast_report(self.sast_report)
        self.workspace.initialize_alert_corpus(self.sast_report)

    @property
    def running(self) -> bool:
        return self._running

    def iter_other_clients(self, cli_id: bytes) -> Generator[PastisClient, None, None]:
        """
        Generator of all clients but the one given in parameter

        :param cli_id: netid of the client to ignore
        :return: Generator of PastisClient object
        """
        for c in self.clients.values():
            if c.netid != cli_id:
                yield c

    def new_uid(self) -> int:
        """
        Generate a new unique id for a client (int)

        :return: int, unique (in an execution)
        :return: int, unique (in an execution)
        """
        v = self._cur_id
        self._cur_id += 1
        return v

    def _register_all(self):
        self.register_seed_callback(self.seed_received)
        self.register_hello_callback(self.hello_received)
        self.register_log_callback(self.log_received)
        self.register_telemetry_callback(self.telemetry_received)
        self.register_stop_coverage_callback(self.stop_coverage_received)
        self.register_data_callback(self.data_received)

    def get_client(self, cli_id: bytes) -> Optional[PastisClient]:
        cli = self.clients.get(cli_id)
        if not cli:
            logging.warning(f"client '{cli_id}' unknown (send stop)")
            self.send_stop(cli_id)
        return cli

    def kick_client(self, cli_id: bytes) -> None:
        cli = self.clients.pop(cli_id)  # pop it from client list
        logging.info(f"kick client: {cli.strid}")
        self.send_stop(cli_id)

    def seed_received(self, cli_id: bytes, typ: SeedType, seed: bytes):
        cli = self.get_client(cli_id)
        if not cli:
            return
        is_new = seed not in self._seed_pool
        h = md5(seed).hexdigest()

        # Show log message and save seed to file
        self.statmanager.update_seed_stat(cli, typ)  # Add info only if new
        cli.log(LogLevel.INFO, f"seed {h} [{self._colored_seed_type(typ)}][{self._colored_seed_newness(is_new)}]")
        cli.add_own_seed(seed)  # Add seed in client's seed
        fname = self.write_seed(typ, cli.strid, seed) # Write seed to file

        if is_new:
            self._seed_pool[seed] = typ  # Save it in the local pool

            if self._coverage_manager:
                sp = fname.split("_")
                covi = ClientInput(seed, "", f"{sp[0]}_{sp[1]}", sp[2], h, fname, typ, cli.netid, cli.strid, "GRANTED", "", -1, [])
                self._coverage_manager.push_input(covi)

            if not self.filter_inputs:  # If seed are not filtered send it right away
                self.seed_granted(cli.netid, typ, seed)
        else:
            logging.debug(f"receive duplicate seed {h} by {cli.strid}")


    def seed_granted(self, cli_id: bytes, typ: SeedType, seed: bytes):
        # Iterate on all clients and send it to whomever never received it
        if self.broker_mode == BrokingMode.FULL:
            self.send_seed_to_all_others(cli_id, typ, seed)

        if self.is_proxied:  # Forward it to the proxy
            self._proxy.send_seed(typ, seed)

    def send_seed_to_all_others(self, origin_id: bytes, typ: SeedType, seed: bytes) -> None:
        for c in self.iter_other_clients(origin_id):
            if c.is_new_seed(seed):
                self.send_seed(c.netid, typ, seed)  # send the seed to the client
                c.add_peer_seed(seed)  # Add it in its list of seed

    def add_seed_file(self, file: PathLike, initial: bool = False) -> None:
        p = Path(file)
        logging.info(f"Add seed {p.name} in pool")
        # Save seed in the workspace
        self.workspace.save_seed_file(SeedType.INPUT, p, initial)

        seed = p.read_bytes()
        self._seed_pool[seed] = SeedType.INPUT
        if initial:
            self._init_seed_pool[seed] = SeedType.INPUT

    def write_seed(self, typ: SeedType, cli_id: str, seed: bytes) -> str:
        fname = self.mk_input_name(cli_id, seed)
        self.workspace.save_seed(typ, fname, seed)
        return fname

    def mk_input_name(self, cli_id: str, seed: bytes) -> str:
        t = time.strftime("%Y-%m-%d_%H:%M:%S", time.localtime())
        elapsed = str(datetime.timedelta(seconds=time.time() - self._start_time)).replace(" day, ", "d:").replace(" days, ", "d:")
        return f"{t}_{elapsed}_{cli_id}_{md5(seed).hexdigest()}.cov"

    def hello_received(self, cli_id: bytes, engines: List[FuzzingEngineInfo], arch: Arch, cpus: int, memory: int, hostname: str, platform: Platform):
        uid = self.new_uid()
        client = PastisClient(uid, cli_id, engines, arch, cpus, memory, hostname, platform)
        logging.info(f"[{client.strid}] [HELLO] Name:{hostname} Arch:{arch.name} engines:{[x.name for x in engines]} (cpu:{cpus}, mem:{memory})")
        self.clients[client.netid] = client

        # Load engines if they are not (lazy loading)
        for eng in engines:
            if eng.name not in self.engines:
                self.load_engine_addon(eng.pymodule)

        if self.running:  # A client is coming in the middle of a session
            if self._startup_quorum:
                self._current_quorum += 1
                if self._current_quorum >= self._startup_quorum:
                    logging.info(
                        f"client number quorum ({self._current_quorum}/{self._startup_quorum}) reached, start all of them")
                    self.start_pending_clients()  # start all pending clients (including the one triggering this func)
                else:
                    # Put the client on-hold to wait for quorum
                    logging.info(f"client {client.strid} on-hold, wait quorum ({self._current_quorum}/{self._startup_quorum})")
            else:  # there is no quorum so start immediately
                self.start_client_and_send_corpus(client)
        else:
            pass  # Client connection is kept in clients dict for later

    def _transmit_pool(self, client, pool) -> None:
        for seed, typ in pool.items():
            self.send_seed(client.netid, typ, seed)  # necessarily a new seed
            client.add_peer_seed(seed)  # Add it in its list of seed

    def log_received(self, cli_id: bytes, level: LogLevel, message: str):
        client = self.get_client(cli_id)
        if not client:
            return
        client.log(level, message)

    def telemetry_received(self, cli_id: bytes,
                           _: State = None, exec_per_sec: int = None, total_exec: int = None,
                           cycle: int = None, timeout: int = None, coverage_block: int = None, coverage_edge: int = None,
                           coverage_path: int = None, last_cov_update: int = None):
        client = self.get_client(cli_id)
        if not client:
            return
        # NOTE: ignore state (shall we do something of it?)
        args = [('-' if not x else x) for x in [exec_per_sec, total_exec, cycle, timeout, coverage_block, coverage_edge, coverage_path, last_cov_update]]
        client.log(LogLevel.INFO, "exec/s:{} tot_exec:{} cycle:{} To:{} CovB:{} CovE:{} CovP:{} last_up:{}".format(*args))

        # Update all stats in the stat manager (for later UI update)
        self.statmanager.set_exec_per_sec(client, exec_per_sec)
        self.statmanager.set_total_exec(client, total_exec)
        self.statmanager.set_cycle(client, cycle)
        self.statmanager.set_timeout(client, timeout)
        self.statmanager.set_coverage_block(client, coverage_block)
        self.statmanager.set_coverage_edge(client, coverage_edge)
        self.statmanager.set_coverage_path(client, coverage_path)
        self.statmanager.set_last_coverage_update(client, last_cov_update)
        self.statmanager.update_telemetry_client(client)
        # NOTE: Send an update signal for future UI ?

    def stop_coverage_received(self, cli_id: bytes):
        client = self.get_client(cli_id)
        if not client:
            return

        if self.ck_mode == CheckMode.ALERT_ONE:  # Start it on another target
            addr = client.target
            s = "validated " if client.target_validated else "is stuck for "
            logging.info(f"[{client.strid}] [STOP_COVERAGE] client {s} 0x{addr:x} address (launch another target)")
            clis = self._slicing_ongoing[client.package_name].pop(addr)  # pop the address definitely
            self.relaunch_clients(clis)  # relaunch all clients on a new target
        else:
            logging.info(f"[{client.strid}] [STOP_COVERAGE]: restart client")
            self.relaunch_clients([client])  # restart the client

    def data_received(self,  cli_id: bytes, data: str):
        client = self.get_client(cli_id)
        if not client:
            return
        first_cov, first_val = False, False

        if not self.sast_report:
            logging.warning("Data received while no SAST report is loaded (drop data)")
            return

        alert_data = AlertData.from_json(data)
        alert = self.sast_report.alerts[alert_data.id]

        if not alert.validated and alert_data.validated:
            logging.info(f"[{client.strid}] First to validate {alert}")
            alert.validated = alert_data.validated
            self.sast_report.write_csv(self.workspace.csv_result_file)  # Update CSV to keep it updated
            first_val = True

            if self.ck_mode == CheckMode.ALERT_ONE:
                client.target_validated = True
                # Note: Do not relaunch the client now, wait for him to send stop_coverage

        # Need to be after validated because alert.covered = True (will also set validated)
        if not alert.covered and alert_data.covered:
            logging.info(f"[{client.strid}] First to cover {alert}")
            alert.covered = alert_data.covered
            self.sast_report.write_csv(self.workspace.csv_result_file)  # Update CSV to keep it updated
            first_cov = True

        # Update clients of and stats
        client.add_covered_alert(alert.id, alert.covered, first_cov, alert.validated, first_val)

        # Save systematically the AlertData received
        self._save_alert_seed(client, alert_data)  # Also save seed in separate folder

        if first_cov or first_val:
            cov, val, tot = self.sast_report.get_stats()
            logging.info(f"SAST results updated: defaults: [covered:{cov}/{tot}] [validated:{val}/{tot}]")

        # If all alerts are validated send a stop to everyone
        if self.sast_report.all_alerts_validated():
            self.stop_broker()

    def relaunch_clients(self, clients):
        for cli in clients:
            logging.info(
                f"Launch client {cli.strid} as its targeting an address that has just been validated")
            self.start_client_and_send_corpus(cli)

    def stop_broker(self):
        for client in self.clients.values():
            logging.info(f"Send stop to {client.strid}")
            self.send_stop(client.netid)
        self._stop = True

        # Stop coverage manager if any
        if self._coverage_manager:
            self._coverage_manager.stop()

        # Call the statmanager to wrap-up values
        self.statmanager.post_execution(list(self.clients.values()), self.workspace)

        if self.is_proxied:
            self._proxy.stop()

        if self.sast_report:  # If a SAST report was loaded
            # Write the final CSV in the workspace
            self.sast_report.write_csv(self.workspace.csv_result_file)
            # And also re-write the Klocwork report (that also contains resutls)
            self.sast_report.write(self.workspace.sast_report_file)

    def start_pending_clients(self) -> None:
        """
        Start clients that connected but for which we did not yet send a response.
        """
        # Send the start message to all clients (already connected)
        for c in self.clients.values():
            if not c.is_running():
                self.start_client_and_send_corpus(c)

    def start_client_and_send_corpus(self, client: PastisClient) -> None:
        self.start_client(client)
        # Iterate all the seed pool and send it to the client
        if self.broker_mode == BrokingMode.FULL:
            self._transmit_pool(client, self._seed_pool)
        elif self.broker_mode == BrokingMode.NO_TRANSMIT:
            self._transmit_pool(client, self._init_seed_pool)

    def start_client(self, client: PastisClient):
        engine = None
        exmode = ExecMode.SINGLE_EXEC
        fuzzmode = FuzzMode.INSTRUMENTED
        engine_args = None
        package = covmode = fuzzmod = None
        engines = Counter({e: 0 for e in self.engines})
        engines.update(c.engine.NAME for c in self.clients.values() if c.is_running())  # Count instances of each engine running
        for eng, _ in engines.most_common()[::-1]:
            eng_desc = self.engines[eng]
            # If the engine is not supported by the client continue
            if not client.is_supported_engine(eng_desc):
                continue

            # Try finding a suitable binary for the current engine and the client arch
            programs: List[BinaryPackage] = self.programs.get((client.platform, client.arch))
            package = None
            exmode = None
            fuzzmod = FuzzMode.AUTO
            for p in programs:
                res, xmod, fmod = eng_desc.accept_file(p.executable_path)  # Iterate all files on that engine descriptor to check it accept it
                if res:
                    if exmode:
                        if exmode == ExecMode.SINGLE_EXEC and xmod == ExecMode.PERSISTENT:  # persistent supersede single_exec
                            package, exmode, fuzzmod = p, xmod, fmod
                        else:
                            if fuzzmod == FuzzMode.BINARY_ONLY and fmod == FuzzMode.INSTRUMENTED:  # instrumented supersede binary only
                                package, exmode, fuzzmod = p, xmod, fmod
                            else:
                                pass  # Program is suitable but we already had found a PERSISTENT one
                    else:
                        package, exmode, fuzzmod = p, xmod, fmod  # first suitable program found

            if not package:  # If still no program was found continue iterating engines
                continue

            # Valid engine and suitable program found
            engine = eng_desc

            # Find a configuration to use for that engine
            engine_args = self._find_configuration(engine)
            covmode = self._find_coverage_mode(engine, engine_args)

            # If got here a suitable program has been found just break loop
            break

        if engine is None or package is None:
            logging.critical(f"No suitable engine or program was found for client {client.strid} {client.engines}")
            return

        if self.ck_mode == CheckMode.ALERT_ONE:
            if package.is_quokka():
                targets = self._slicing_ongoing[package.name]
                sorted_targets = sorted(targets, key=lambda k: len(targets[k]), reverse=False)  # sort alert addresses by number of client instances working on it
                if sorted_targets:
                    addr = sorted_targets[0]
                    targets[addr].append(client)
                    client.target = addr   # keep the target on which the client is working on
                    client.target_validated = False
                    # Now twist the config to transmit it to the client
                    engine_args = engine.config_class.new() if engine_args is None else engine_args
                    engine_args.set_target(addr)
                    logging.info(f"will start client {client.strid} to target 0x{addr:x}")
                else:
                    logging.error(f"No alert target for binary package {package.name}")
            else:
                logging.error(f"In mode {self.ck_mode} but the binary package is not a QBinExport")
                return

        # Update internal client info and send him the message
        engine_args_str = engine_args.to_str() if engine_args else ""
        logging.info(f"send start client {client.id}: {package.name} [{engine.NAME}, {covmode.name}, {fuzzmod.name}, {exmode.name}]")
        client.set_running(package.name, engine, covmode, exmode, self.ck_mode, engine_args_str)
        client.configure_logger(self.workspace.log_directory, random.choice(COLORS))  # Assign custom color client

        self.send_start(client.netid,
                        package.name,
                        package.executable_path if package.is_standalone() else package.make_package(),
                        self.argv,
                        exmode,
                        fuzzmod,
                        self.ck_mode,
                        covmode,
                        FuzzingEngineInfo(engine.NAME, engine.VERSION, ""),
                        engine_args_str,
                        self.inject,
                        self.sast_report.to_json() if self.sast_report else b"")

    def _find_configuration(self, engine: FuzzingEngineDescriptor) -> Optional[EngineConfiguration]:
        """
        Find a configuration for the engine. It will iterate all configuration
        available or automatically balance de configuration if there is multiple of
        them
        :param engine:
        :return:
        """
        confs = self.engines_args[engine.NAME]
        if confs:
            if len(confs) == 1:
                return confs[0]
            else:
                conf = confs.pop(0)
                confs.append(conf)  # Rotate the configuration
                return conf
        else:
            return None

    def _find_coverage_mode(self, engine: FuzzingEngineDescriptor, conf: Union[EngineConfiguration]) -> CoverageMode:

        # Get coverage modes supported by the engine
        supported_mods = engine.supported_coverage_strategies()

        # Now that program have been found select coverage strategy
        if len(supported_mods) > 1:

            if conf:  # In the case we wanted to launch engines with specific configuration
                return conf.get_coverage_mode()    # return the appropriate coverage mode in that configuration

            else:
                # auto-balance the CoverageMode instances
                covs = Counter({c: 0 for c in supported_mods})
                # Count how many times each coverage strategies have been launched
                covs.update(x.coverage_mode for x in self.clients.values() if x.is_running() and x.engine == engine)

                if sum(covs.values()) == 0:  # No configuration has been triggered yet
                    if CoverageMode.EDGE in supported_mods:  # launch preferably edge first
                        return CoverageMode.EDGE
                    else:
                        return supported_mods[0]  # Return first supported mode

                reverted = {v: k for k, v in covs.items()}  # Revert dictionnary to have freq -> covmodes
                return reverted[min(reverted)]  # Select mode if least instances

        else:
            return supported_mods[0]

    def start(self, running: bool = True):
        super(PastisBroker, self).start()  # Start the listening thread
        self._start_time = time.time()
        self._running = running
        self.workspace.status = WorkspaceStatus.RUNNING
        logging.info("start broking")

        if self._coverage_manager:  # if it has been instanciated start it
            self._coverage_manager.start()
            for seed in self._init_seed_pool.keys():  # Push initial corpus to set baseline coverage
                fname = self.mk_input_name("INITIAL", seed)
                sp = fname.split("_")
                covi = ClientInput(seed, "", f"{sp[0]}_{sp[1]}", sp[2], sp[4], fname, SeedType.INPUT, b"INITIAL", "INITIAL", "GRANTED", "", -1, [])
                self._coverage_manager.push_input(covi)

        if self.is_proxied and self._proxy_cli:
            self._running = False  # disable running wait start broker
            self._proxy.send_hello([self._proxy_cli])
        else:
            # Send the start message to all clients (already connected)
            if self._running:  # If we want to run now (cmdline mode)
                self.start_pending_clients()

    def run(self, timeout: int = None):
        self.start()
        last_t = time.time()

        # Start infinite loop
        try:
            while True:
                time.sleep(0.05)
                t = time.time()

                # Check if the campaign have to be stopped
                if timeout is not None:
                    if t > (self._start_time + timeout):
                        logging.info("Campaign timeout reached, stop campaign.")
                        self._stop = True

                if t > (last_t + 60):  # only check every minute
                    last_t = t
                    if not self._check_memory_usage():
                        # The machine starts being overloaded
                        # For security kill triton instance
                        for cli in list(self.clients.values()):
                            if cli.engine.SHORT_NAME == "TT":  # is triton
                                self.kick_client(cli.netid)

                # if inputs are filtered. Get granted inputs and forward them to appropriate clients
                if self.filter_inputs:
                    for item in self._coverage_manager.iter_granted_inputs():
                        self.seed_granted(item.fuzzer_id, item.seed_status, item.content)

                # Check if we received the start signal from the proxy-master
                if self._proxy_start_signal:
                    self._proxy_start_signal = False
                    self.start_pending_clients()

                # Check if there are seed coming from the proxy-master to forward to clients
                if not self._proxy_seed_queue.empty():
                    try:
                        while True:
                            origin, typ, seed = self._proxy_seed_queue.get_nowait()
                            self.send_seed_to_all_others(origin, typ, seed)
                    except queue.Empty:
                        pass

                if self._stop:
                    logging.info("broker terminate")
                    break
        except KeyboardInterrupt:
            logging.info("stop required (Ctrl+C)")
        self.workspace.status = WorkspaceStatus.FINISHED
        self.stop_broker()

    def _find_binaries(self, binaries_dir) -> None:
        """
        Iterate the whole directory to find suitables binaries in the
        various architecture, and compiled for the various engines.

        :param binaries_dir: directory containing the various binaries
        :return: None
        """
        d = Path(binaries_dir)
        for file in d.iterdir():
            if file.is_file():
                try:
                    pkg = BinaryPackage.auto(file)  # try creating a package
                except ValueError:  # if not an executable
                    continue

                if pkg is None:
                    continue

                # Check that we can find a Quokka file associated otherwise reject it
                if self.ck_mode == CheckMode.ALERT_ONE:
                    if pkg.is_quokka():
                        try:
                            # Instanciate the Quokka Program and monkey patch object
                            quokka_prog = QuokkaProgram(pkg.quokka, pkg.executable_path)
                            f = quokka_prog.get_function("__sast_alert_placeholder")
                            self._slicing_ongoing[file.name] = {x: [] for x in quokka_prog.get_caller_instructions(f)}
                        except ValueError:
                            logging.warning(f"can't find placeholder file in {file.name}, thus ignores it.")
                            continue
                    else:
                        logging.warning(f"{file.name} executable found but no QBinExport file associated (ignores it)")
                        continue

                logging.info(f"new binary detected [{pkg.platform.name}, {pkg.arch.name}]: {file}")

                # Add it in the internal structure
                data = (pkg.platform, pkg.arch)
                data2 = (Platform.ANY, pkg.arch)
                if data not in self.programs:
                    self.programs[data] = []
                if data2 not in self.programs:
                    self.programs[data2] = []
                self.programs[data].append(pkg)
                self.programs[data2].append(pkg)  # Also add an entry for any platform

    def _load_workspace(self):
        """ Load all the seeds in the workspace"""
        for typ in list(SeedType):  # iter seed types: input, crash, hang..
            for file in self.workspace.iter_corpus_directory(typ):
                logging.debug(f"Load seed in pool: {file.name}")
                content = file.read_bytes()
                self._seed_pool[content] = typ
        # TODO: Also dumping the current state to a file in case
        # TODO: of exit. And being able to reload it. (not to resend all seeds to clients)

    def add_engine_configuration(self, name: str, config_file: PathLike):
        if name in self.engines_args:
            engine = self.engines[name]
            conf = engine.config_class.from_file(config_file)
            self.engines_args[name].append(conf)
        else:
            logging.error(f"can't find engine {name} (shall preload it to add a configuration)")

    def _configure_logging(self):
        hldr = logging.FileHandler(self.workspace.broker_log_file)
        hldr.setLevel(logging.root.level)
        hldr.setFormatter(logging.Formatter("%(asctime)s [%(name)s] [%(levelname)s]: %(message)s"))
        logging.root.addHandler(hldr)

    def _colored_seed_type(self, typ: SeedType) -> str:
        mapper = {SeedType.INPUT: Bcolors.OKBLUE,
                  SeedType.HANG: Bcolors.WARNING,
                  SeedType.CRASH: Bcolors.FAIL}
        return mapper[typ]+typ.name+Bcolors.ENDC

    def _colored_seed_newness(self, is_new: bool) -> str:
        col, text = {True: (Bcolors.OKGREEN, "NEW"),
                     False: (Bcolors.WARNING, "DUP")}[is_new]
        return col+text+Bcolors.ENDC

    def _save_alert_seed(self, from_cli: PastisClient, alert: AlertData):
        t = time.strftime("%Y-%m-%d_%H:%M:%S", time.localtime())
        elapsed = str(datetime.timedelta(seconds=time.time()-self._start_time)).replace(" day, ", "d:").replace(" days, ", "d:")
        fname = f"{t}_{elapsed}_{from_cli.strid}_{md5(alert.seed).hexdigest()}-{'CRASH' if alert.validated else 'COVERAGE'}.cov"
        logging.debug(f"Save alert  [{alert.id}] file: {fname}")
        self.workspace.save_alert_seed(alert.id, fname, alert.seed)

    def _check_memory_usage(self) -> bool:
        mem = psutil.virtual_memory()
        logging.info(f"RAM usage: {mem.percent:.2f}%")
        if mem.percent >= self._threshold:
            logging.warning(f"Threshold reached: {mem.percent}%")
            return False
        return True

    @property
    def is_proxied(self) -> bool:
        """
        Get if the broker as a proxy to another broker.
        """
        return self._proxy is not None

    def set_proxy(self, ip: str, port: int, py_module: str) -> bool:
        self._proxy = ClientAgent()

        # Load the engine info to impersonate
        desc = load_engine_descriptor(py_module)
        if desc is None:
            logging.error("Cannot load FuzzingEngineDescriptor")
            return False
        else:
            self._proxy_cli = FuzzingEngineInfo(desc.NAME, desc.VERSION, py_module)

        self._proxy.register_start_callback(self._proxy_start_received)
        self._proxy.register_seed_callback(self._proxy_seed_received)
        self._proxy.register_stop_callback(self._proxy_stop_received)

        logging.info(f"connect to broker {ip}:{port}")
        self._proxy.connect(ip, port)
        self._proxy.start()

    def _proxy_start_received(self, fname: str, binary: bytes, engine: FuzzingEngineInfo, exmode: ExecMode,
                              fuzzmode: FuzzMode, chkmode: CheckMode, covmode: CoverageMode, seed_inj: SeedInjectLoc,
                              engine_args: str, argv: List[str], sast_report: str = None):
        # FIXME: Use parameters received
        logging.info("[PROXY] start received !")
        self._running = True
        # if self._running:
        #     self.start_pending_clients()

    def _proxy_seed_received(self, typ: SeedType, seed: bytes):
        # Forward the seed to underlying clients
        logging.info(f"[PROXY] seed {typ.name} received forward to agents")

        # Save the seed locally
        self.write_seed(typ, "PROXY", seed)
        self._seed_pool[seed] = typ  # add it to the pool
        self._init_seed_pool[seed] = typ  # also consider it as initial corpus

        # Forward it to all clients
        self._proxy_seed_queue.put((b"PROXY", typ, seed))
        # self.send_seed_to_all_others(b"PROXY", typ, seed)

    def _proxy_stop_received(self):
        logging.info(f"[PROXY] stop received!")
        self._stop = True



#!/usr/bin/env python3
import time
import logging
import tempfile
import os
import subprocess
from typing import Generator
from pathlib import Path
import queue
import csv
from dataclasses import dataclass
from threading import Thread
from multiprocessing import Queue, Manager
from multiprocessing.pool import Pool

from libpastis.types import SeedType, SeedInjectLoc

# tritondse imports
from tritondse import GlobalCoverage, CoverageSingleRun, CoverageStrategy, BranchSolvingStrategy
from tritondse.trace import QBDITrace, TraceException

from pastisbroker.utils import Bcolors, mk_color


@dataclass
class ClientInput:
    content: bytes          # Content of the input
    log_time: str           # Time the log has been generated
    recv_time: str          # Time the input has been received
    elapsed: str            # Elapsed time since the begining
    hash: str               # Input hash
    path: str               # Input file path
    seed_status: SeedType   # Status of the seed
    fuzzer_id: bytes        # Fuzzer ID
    fuzzer_name: str        # Fuzzer name
    broker_status: str      # Status in: DUPLICATE, DROPPED, GRANTED
    replay_status: str      # Status in: OK, TRACE_EXCEPTION, FAIL
    replay_time: float      # Time taken for the replay
    new_coverage: list[tuple[int, int]]  # New items covered


class CoverageManager(object):

    ARGV_PLACEHOLDER = "@@"
    STRATEGY = CoverageStrategy.EDGE

    def __init__(self, pool_size: int, filter: bool, program: str, args: list[str], inj_loc: SeedInjectLoc, stream_file: str = ""):
        # Base info for replay
        self.pool_size = pool_size
        self.filter_enabled = filter
        self.program = str(program)
        self.args = args
        self.inj_loc = inj_loc

        # Coverage and messaging attributes
        self._coverage = GlobalCoverage(self.STRATEGY, BranchSolvingStrategy.ALL_NOT_COVERED)
        self._manager = Manager()
        self.input_queue = self._manager.Queue()
        self.cov_queue = self._manager.Queue()
        self.granted_queue = self._manager.Queue()

        # Pool of workers
        self.pool = Pool(self.pool_size)
        self._running = False
        self.cov_worker = Thread(name="[coverage_worker]", target=self.coverage_worker)

        # stats
        self.seeds_accepted, self.seeds_submitted = 0, 0
        self.cli_stats = {}

        # Streaming
        if stream_file:
            self.stream_file = open(stream_file, "a")
            self.csv = csv.writer(self.stream_file)
        else:
            self.stream_file, self.csv = None, None


    def start(self) -> None:
        """
        Start all the workers
        """
        # First start the coverage worker
        self._running = True
        self.cov_worker.start()
        logging.info("Starting coverage manager")

        for work_id in range(self.pool_size):
            self.pool.apply_async(self.worker, (self.input_queue, self.cov_queue, self.program, self.args, self.inj_loc))

    def stop(self) -> None:
        self._running = False
        self.cov_worker.join()
        self.pool.terminate()

    def push_input(self, cli_input: ClientInput) -> None:
        """ Push the input in the """
        cli_input.log_time = time.strftime("%Y-%m-%d_%H:%M:%S", time.localtime())
        # logging.info(f"push input {str(cli_input)[:50]}")

        # Update stats
        self.seeds_submitted += 1
        if cli_input.fuzzer_id in self.cli_stats:
            self.cli_stats[cli_input.fuzzer_id][0] += 1
        else:
            self.cli_stats[cli_input.fuzzer_id] = [1, 0]

        self.input_queue.put(cli_input)

    def iter_granted_inputs(self) -> Generator[ClientInput, None, None]:
        try:
            while True:
                yield self.granted_queue.get_nowait()
        except queue.Empty:
            pass

    @staticmethod
    def worker_sleep(q, n) -> None:
        """
        worker thread that unstack inputs and replay them.
        """
        time.sleep(n)
        q.put_nowait(n)
        return n

    def add_item_coverage_stream(self, item: ClientInput) -> None:
        if self.stream_file:  # Stream enabled
            self.csv.writerow([
                item.log_time,
                item.recv_time,
                item.elapsed,
                item.hash,
                item.path,
                item.seed_status.name,
                item.fuzzer_name,
                item.broker_status,
                item.replay_status,
                f"{item.replay_time:.2f}",
                item.new_coverage
            ])
            self.stream_file.flush()

    def coverage_worker(self):
        while self._running:
            try:
                item, cov_file = self.cov_queue.get(timeout=0.5)
                # logging.info("Coverage worker fetch item")
                new_items = []
                try:
                    coverage: CoverageSingleRun = QBDITrace.from_file(cov_file).coverage
                    if self._coverage.improve_coverage(coverage):
                        self.cli_stats[item.fuzzer_id][1] += 1  # input accepted
                        self.seeds_accepted += 1

                        # Get newly covered items (and put them in the stream queue

                        new_items = coverage.difference(self._coverage)

                        item.new_coverage = list(new_items)

                        # logging.info(f"seed {item.hash}  ({item.fuzzer_name}) [replay:{}][status:{}] ({len(new_items)} new edges)")

                        # Update the global coverage
                        self._coverage.merge(coverage)

                        if item.fuzzer_name != "INITIAL":  # if not initial corpus and granted
                            self.granted_queue.put(item)

                    else:
                        item.broker_status = "DROPPED" if self.filter_enabled else "GRANTED"
                        # logging.info(f"seed {item.hash} ({item.seed_status.name}) of {item.fuzzer_name} rejected (do not improve coverage)")

                    # Remove the coverage file
                    os.unlink(cov_file)
                except FileNotFoundError:
                    if item.seed_status == SeedType.INPUT:
                        logging.warning(f"seed {item.hash}({item.seed_status}) can't load coverage file (maybe had crashed?)")
                    else:
                        logging.info(f"seed {item.hash}({item.seed_status}) cannot get coverage (normal..)")
                    # Grant input
                    self.seeds_accepted += 1
                    self.granted_queue.put(item)

                logging.info(f"seed {item.hash} ({item.fuzzer_name}) [replay:{self.mk_rpl_status(item.replay_status)}][status:{self.mk_broker_status(item.broker_status, bool(new_items))}] ({len(new_items)} new edges)")
                # Regardless if it was a success or not log it
                self.add_item_coverage_stream(item)
            except queue.Empty:
                pass
            except KeyboardInterrupt:
                self._running = False
                logging.info("coverage worker stop")
                break

    @staticmethod
    def mk_rpl_status(status: str) -> str:
        if status == "SUCCESS":
            return mk_color(status, Bcolors.OKGREEN)
        else:
            return mk_color(status, Bcolors.FAIL)


    @staticmethod
    def mk_broker_status(status: str, new_items: bool) -> str:
        if status == "GRANTED":
            return mk_color(status, Bcolors.OKGREEN if new_items else Bcolors.WARNING)
        elif status == "DROPPED":
            return mk_color(status, Bcolors.WARNING)
        else:
            return mk_color(status, Bcolors.FAIL)

    @staticmethod
    def worker(input_queue: Queue, cov_queue: Queue, program: str, argv: list[str], seed_inj: SeedInjectLoc) -> None:
        """
        worker thread that unstack inputs and replay them.
        """
        tmpfile = Path(tempfile.mktemp(suffix=f"{os.getpid()}.input"))
        pid = os.getpid()
        try:
            while True:
                item: ClientInput = input_queue.get()
                # logging.debug(f"Worker {os.getpid()} fetch: {str(item)[:50]}")
                # Write inputs in our tempfile
                tmpfile.write_bytes(item.content)

                # Create to coverage file
                cov_file = tempfile.mktemp(f"_{item.hash}.cov")

                # Adjust injection location before calling QBDITrace
                cur_argv = argv[:]
                if seed_inj == SeedInjectLoc.ARGV:  # Try to replace the placeholder with filename
                    try:
                        # Replace 'input_file' in argv with the temporary file name created
                        idx = cur_argv.index(CoverageManager.ARGV_PLACEHOLDER)
                        cur_argv[idx] = str(tmpfile)
                    except ValueError as e:
                        logging.error(f"seed injection {seed_inj.name} but can't find '@@' on program argv: {argv}: {e}")
                        continue

                try:
                    # Run the seed
                    t0 = time.time()
                    if QBDITrace.run(CoverageManager.STRATEGY,
                                     program,
                                     cur_argv,  # argv[1:] if len(argv) > 1 else [],
                                     output_path=str(cov_file),
                                     stdin_file=str(tmpfile) if seed_inj == SeedInjectLoc.STDIN else None,
                                     cwd=Path(program).parent,
                                     timeout=60):
                        item.replay_time = time.time() - t0
                        item.replay_status = "SUCCESS"
                        # logging.info(f"[worker-{pid}] replaying {item.hash} sucessful")
                    else:
                        item.replay_status = "FAIL_NO_COV"
                        logging.warning("Cannot load the coverage file generated (maybe had crashed?)")
                except TraceException:
                    item.replay_status = "FAIL_TIMEOUT"
                    logging.warning('Timeout hit, while trying to re-run the seed')

                # Add it to the coverage queue (even if it failed
                cov_queue.put((item, cov_file))
        except KeyboardInterrupt:
            pass
            # logging.info(f"replay worker {os.getpid()}, stops (keyboard interrupt)")



import logging
import os
import subprocess
import re
import signal
import time
from pathlib import Path
from typing import Optional
from libpastis.types import ExecMode, FuzzMode

# Local imports
from .workspace import Workspace


class HonggfuzzNotFound(Exception):
    """ Issue raised on """
    pass


class HonggfuzzProcess:

    HFUZZ_ENV_VAR = "HFUZZ_PATH"
    HFUZZ_THREADS_VAR = "HFUZZ_THREADS"
    BINARY = "honggfuzz"
    STAT_FILE = "statsfile.log"
    VERSION = "2.1"

    def __init__(self, path: str = None):
        path = os.environ.get(self.HFUZZ_ENV_VAR) if path is None else path
        if path is None:
            raise Exception("Invalid Honggfuzz path provided")

        path = Path(path)
        if not path.exists():
            raise Exception('Invalid HFUZZ_PATH path!')
        elif path.is_file() and path.name == self.BINARY:
            self.__path = path
        elif path.is_dir():
            self.__path = Path(path) / self.BINARY
            if not path.exists():
                raise Exception("Can't find honggfuzz in HFUZZ_PATH path!")

        self._threads = os.environ.get(self.HFUZZ_THREADS_VAR)

        self.__process = None

    def start(self, target: str, target_arguments: list[str], workspace: Workspace, exmode: ExecMode, fuzzmode: FuzzMode,
              stdin: bool, engine_args: str, dictionary: Optional[str] = None) -> bool:
        if not stdin:
            if "@@" in target_arguments:  # Change '@@' for ___FILE___
                idx = target_arguments.index("@@")
                target_arguments[idx] = "___FILE___"
            else:
                if "___FILE___" not in target_arguments:
                    logging.error(f"seed provided via ARGV but can't find '@@'/___FILE___ on program argv")
                    return False

        # Build target command line.
        target_cmdline = f"{target} {' '.join(target_arguments)}"

        HFQBDI_LIB_PATH = os.getenv('HFQBDI_LIB_PATH')

        if fuzzmode == FuzzMode.BINARY_ONLY and HFQBDI_LIB_PATH is None:
            logging.error(f"target in BINARY_ONLY but can't find HFQBDI_LIB_PATH")
            return False

        # Build fuzzer arguments.
        hfuzz_arguments = ' '.join([
            f"--statsfile {workspace.stats_file}",
            f"--stdin_input" if stdin else "",
            f"--persistent" if exmode == ExecMode.PERSISTENT or fuzzmode == FuzzMode.BINARY_ONLY else "",
            f"--env HFQBDI_FS=1" if fuzzmode == FuzzMode.BINARY_ONLY else "",
            f"--env LD_LIBRARY_PATH={HFQBDI_LIB_PATH}" if fuzzmode == FuzzMode.BINARY_ONLY else "",
            f"--env LD_PRELOAD={HFQBDI_LIB_PATH}/libHFQBDIpreload.so" if fuzzmode == FuzzMode.BINARY_ONLY else "",
            f"--env LD_BIND_NOW=1" if fuzzmode == FuzzMode.BINARY_ONLY else "",
            re.sub(r"\s", " ", engine_args),  # Any arguments coming right from the broker (remove \r\n)
            f"--logfile logfile.log",
            f"--input {workspace.input_dir}",
            f"--dynamic_input {workspace.dynamic_input_dir}",
            f"--output {workspace.corpus_dir}",
            f"--crashdir {workspace.crash_dir}",
            f"--workspace {workspace.root_dir}",
            f"--threads {self._threads}" if self._threads else "",
            f"--dict {dictionary}" if dictionary is not None else ""
        ])

        # Build fuzzer command line.
        hfuzz_cmdline = f'{self.__path} {hfuzz_arguments} -- {target_cmdline}'

        logging.info(f"Run Honggfuzz with: {hfuzz_cmdline}")
        logging.debug(f"\tWorkspace: {workspace.root_dir}")

        # Remove empty strings when converting the command to a list.
        command = list(filter(None, hfuzz_cmdline.split(' ')))

        # Create a new fuzzer process and set it apart into a new process group.
        self.__process = subprocess.Popen(command, cwd=str(workspace.root_dir), preexec_fn=os.setsid)

        logging.debug(f'Process pid: {self.__process.pid}')
        return True

    @property
    def instanciated(self):
        return self.__process is not None

    def stop(self):
        if self.__process:
            logging.debug(f'Stopping process with pid: {self.__process.pid}')
            os.killpg(os.getpgid(self.__process.pid), signal.SIGTERM)
        else:
            logging.debug(f"Honggfuzz process seem's already killed")

    def wait(self):
        while not self.instanciated:
            time.sleep(0.1)
        self.__process.wait()

    @staticmethod
    def hfuzz_environ_check() -> bool:
        path = os.environ.get(HonggfuzzProcess.HFUZZ_ENV_VAR)
        if path is None:
            return False
        else:
            return (Path(path) / HonggfuzzProcess.BINARY).exists()



# built-in imports
from enum import Enum, auto
from pathlib import Path
from typing import Generator, Optional
import os
import subprocess
import logging
import time
import re
from datetime import datetime, timedelta

# third-party
from pastisbroker.workspace import Workspace, WorkspaceStatus
from libpastis.types import SeedInjectLoc, SeedType
from tritondse.trace import QBDITrace, TraceException
from tritondse import CoverageStrategy


class ReplayType(Enum):
    qbdi = auto()
    llvm_profile = auto()


class Replayer(object):

    QBDI_REPLAY_DIR = "replays_qbdi"
    LLVMPROFILE_REPLAY_DIR = "replays_llvmprof"
    REPLAY_FAILS_LOG = "replay_fails.log"

    def __init__(self, program: Path, workspace: Path, type: ReplayType, injloc: SeedInjectLoc,
                 stream: bool = False, full: bool = False, timeout: int = 15, *args):
        self.workspace = Workspace(workspace)
        self.type = type
        self.stream = stream
        self._full = full
        self.program = Path(program)
        self._inject_loc = injloc
        self._timeout = timeout
        self._args = list(args)
        self._fails = []
        self._tracing_times = []

        # initiatialize directories
        self._init_directories()

    def _init_directories(self):
        if not self.corpus_replay_dir.exists():
            self.corpus_replay_dir.mkdir()

    @property
    def corpus_replay_dir(self) -> Path:
        if self.type == ReplayType.qbdi:
            return self.workspace.root / self.QBDI_REPLAY_DIR
        else:
            return self.workspace.root / self.LLVMPROFILE_REPLAY_DIR

    def iter(self) -> Generator[Path, None, None]:
        yield from self.workspace.iter_initial_corpus_directory()
        yield from self.workspace.iter_corpus_directory(SeedType.INPUT)

    def replay(self, input: Path) -> bool:
        if self.type == ReplayType.qbdi:
            return self._replay_qbdi(input)
        else:
            return self._replay_llvm_profile(input)

    def _replay_qbdi(self, input: Path) -> bool:
        out_file = self.corpus_replay_dir / (input.name + ".trace")

        if out_file.exists():
            # The trace has already been generated
            return True

        args = self._args[:]

        # If inject on argv try replacing the right argv
        if self._inject_loc == SeedInjectLoc.ARGV:
            if "@@" in args:
                idx = args.index("@@")
                args[idx] = str(input.absolute())

        try:
            t0 = time.time()
            res = QBDITrace.run(CoverageStrategy.EDGE,
                                 str(self.program.absolute()),
                                 args=args,
                                 output_path=str(out_file.absolute()),
                                 stdin_file=input if self._inject_loc == SeedInjectLoc.STDIN else None,
                                 dump_trace=self._full,
                                 cwd=self.program.parent,
                                 timeout=self._timeout)
            self._tracing_times.append(time.time()-t0)
            return res
        except TraceException as e:
            self._fails.append(input)
            return False

    def _replay_llvm_profile(self, input: Path) -> bool:
        pass


    def start(self):
        # TODO: Start monitoring folders (and status file)
        pass

    def save_fails(self):
        with open(self.workspace.root / self.REPLAY_FAILS_LOG, "w") as f:
            f.write("\n".join(str(x) for x in self._fails))

    def print_stats(self):
        def tt(secs):
            return str(timedelta(seconds=int(secs)))
        if not self._tracing_times:
            print("nothing replayed")
            return
        sum_tracing = sum(self._tracing_times)
        mean_replay = sum_tracing / len(self._tracing_times)
        print(f"Tracing: {tt(sum_tracing)} (avg: {mean_replay}s)")



from .agent import FileAgent, BrokerAgent, ClientAgent
from .enginedesc import EngineConfiguration, FuzzingEngineDescriptor
from .package import BinaryPackage
from .sast import SASTAlert, SASTReport

__version__ = "1.0.0"



# built-ins
import time
from typing import Callable, Tuple, List, Union
from enum import Enum
import logging
import threading
from pathlib import Path
import socket

# third-party libs
import zmq
import psutil

# local imports
from libpastis.proto import InputSeedMsg, StartMsg, StopMsg, HelloMsg, LogMsg, \
                            TelemetryMsg, StopCoverageCriteria, DataMsg, EnvelopeMsg
from libpastis.types import SeedType, Arch, FuzzingEngineInfo, PathLike, ExecMode, CheckMode, CoverageMode, SeedInjectLoc, \
                            LogLevel, State, AlertData, Platform, FuzzMode
from libpastis.utils import get_local_architecture, get_local_platform

Message = Union[InputSeedMsg, StartMsg, StopMsg, HelloMsg, LogMsg, TelemetryMsg, StopCoverageCriteria, DataMsg]


class MessageType(Enum):  # Topics in the ZMQ terminology
    """
    Enum encoding the type of the message that can be received.
    """
    HELLO = 'hello_msg'
    # STATE = 1
    START = 'start_msg'
    INPUT_SEED = 'input_msg'
    TELEMETRY = 'telemetry_msg'
    LOG = 'log_msg'
    STOP_COVERAGE_DONE = 'stop_crit_msg'
    STOP = "stop_msg"
    DATA = "data_msg"


class AgentMode(Enum):
    """
    Internal enum identifying whether the agent is running as a broker
    or a client.
    """
    BROKER = 1
    CLIENT = 2


class NetworkAgent(object):
    """
    Base class for network-based PASTIS agents (both clients and servers)
    """
    def __init__(self):
        self.mode = None
        self.ctx = zmq.Context()
        self.socket = None
        self._stop = False
        self._th = None
        self._cbs = {x: [] for x in MessageType}

    def register_callback(self, typ: MessageType, callback: Callable) -> None:
        """
        Register a callback function on a given message type.

        :param typ: type of the message
        :param callback: Callback function taking the protobuf object as parameter
        :return: None
        """
        self._cbs[typ].append(callback)

    def bind(self, port: int = 5555, ip: str = "*") -> None:
        """
        Bind on the given IP and port, to listen incoming messages.

        :param port: listen port
        :param ip: IP, can be "*" to listen on all interfaces
        :return: None
        """
        self.socket = self.ctx.socket(zmq.ROUTER)
        self.socket.RCVTIMEO = 500  # 500 milliseconds
        self.socket.bind(f"tcp://{ip}:{port}")
        self.mode = AgentMode.BROKER

    def connect(self, remote: str = "localhost", port: int = 5555) -> bool:
        """
        Connect to a remote server on the given ``remote`` IP and ``port``.

        :param remote: IP address or DNS
        :param port: port to connect to
        :return: Always true
        """
        self.socket = self.ctx.socket(zmq.DEALER)
        self.socket.RCVTIMEO = 500  # 500 milliseconds
        self.socket.connect(f"tcp://{remote}:{port}")
        self.mode = AgentMode.CLIENT
        return True

    def start(self) -> None:
        """
        Start the listening thread.
        """
        self._th = threading.Thread(name="[LIBPASTIS]", target=self._recv_loop, daemon=True)
        self._th.start()

    def run(self) -> None:
        """
        Run receiving loop in a blocking manner.
        """
        self._recv_loop()

    def stop(self) -> None:
        """
        Stop the listening thread.
        """
        self._stop = True
        if self._th:
            self._th.join()

    def _recv_loop(self):
        #flags = 0 if blocking else zmq.DONTWAIT
        while 1:
            if self._stop:
                return
            try:
                if self.mode == AgentMode.BROKER:
                    uid, data = self.socket.recv_multipart()
                    self.__broker_transfer_to_callback(uid, data)
                else:
                    data = self.socket.recv()
                    self.__client_transfer_to_callback(data)
            except zmq.error.Again:
                pass

    def send_to(self, id: bytes, msg: Message, msg_type: MessageType = None) -> None:
        """
        Send a message to a given client. Only meant to be used when
        running as a server.

        :param id: bytes id of the client
        :param msg: protobuf :py:obj:`Message` object
        :param msg_type: type of the message
        """
        if self.mode == AgentMode.CLIENT:
            logging.error(f"cannot use sento_to() as {AgentMode.CLIENT.name}")
            return
        if msg_type is None:
            msg_type = self.msg_to_type(msg)
        final_msg = EnvelopeMsg()
        getattr(final_msg, msg_type.value).MergeFrom(msg)
        self.socket.send_multipart([id, final_msg.SerializeToString()])

    def send(self, msg: Message, msg_type: MessageType = None) -> None:
        """
        Send a message on the socket (thus to the broker). Should
        only be used as a client (fuzzing agent).

        :param msg: Protobuf message to send
        :param msg_type: Type of the message
        """
        if self.mode == AgentMode.BROKER:
            logging.error(f"cannot use sento() as {AgentMode.BROKER.name}")
            return
        if msg_type is None:
            msg_type = self.msg_to_type(msg)
        final_msg = EnvelopeMsg()
        getattr(final_msg, msg_type.value).CopyFrom(msg)
        self.socket.send(final_msg.SerializeToString())

    @staticmethod
    def msg_to_type(msg: Message) -> MessageType:
        """
        Get the :py:obj:`MessageType` from a protobuf object.

        :param msg: Protobuf message
        :return: message type
        """
        if isinstance(msg, InputSeedMsg):
            return MessageType.INPUT_SEED
        elif isinstance(msg, HelloMsg):
            return MessageType.HELLO
        elif isinstance(msg, TelemetryMsg):
            return MessageType.TELEMETRY
        elif isinstance(msg, LogMsg):
            return MessageType.LOG
        elif isinstance(msg, StopMsg):
            return MessageType.STOP
        elif isinstance(msg, StopCoverageCriteria):
            return MessageType.STOP_COVERAGE_DONE
        elif isinstance(msg, StartMsg):
            return MessageType.START
        elif isinstance(msg, DataMsg):
            return MessageType.DATA
        else:
            logging.error(f"invalid message type: {type(msg)} (cannot find associated topic)")

    def __broker_transfer_to_callback(self, id: bytes, message: bytes):
        try:
            msg = EnvelopeMsg()
            msg.ParseFromString(message)
        except:
            logging.error(f"can't parse message from {id} (len:{len(message)})")
            return
        message, topic = self._unpack_message(msg)
        if topic in [MessageType.START]:
            logging.error(f"Invalid message of type {topic.name} received")
        if not self._cbs[topic]:
            logging.warning(f"[broker] message of type {topic.name} (but no callback)")
        args = self._message_args(topic, message)
        for cb in self._cbs[topic]:
            cb(id, *args)

    def __client_transfer_to_callback(self, message: bytes):
        msg = EnvelopeMsg()
        msg.ParseFromString(message)
        message, topic = self._unpack_message(msg)
        if topic in [MessageType.HELLO, MessageType.TELEMETRY, MessageType.LOG, MessageType.STOP_COVERAGE_DONE]:
            logging.error(f"Invalid message of type {topic.name} received")
        if not self._cbs[topic]:
            logging.warning(f"[agent] message of type {topic.name} (but no callback)")
        args = self._message_args(topic, message)
        for cb in self._cbs[topic]:
            cb(*args)

    def _unpack_message(self, message: EnvelopeMsg) -> Tuple[MessageType, Message]:
        typ = message.WhichOneof('msg')
        return getattr(message, typ), MessageType(typ)

    def _message_args(self, topic: MessageType, msg: Message):
        if topic == MessageType.INPUT_SEED:
            return [SeedType(msg.type), msg.seed]
        elif topic == MessageType.LOG:
            return [LogLevel(msg.level), msg.message]
        elif topic == MessageType.TELEMETRY:
            return [msg.state, msg.exec_per_sec, msg.total_exec, msg.cycle, msg.timeout, msg.coverage_block,
                    msg.coverage_edge, msg.coverage_path, msg.last_cov_update]
        elif topic == MessageType.HELLO:
            engs = [(FuzzingEngineInfo.from_pb(x)) for x in msg.engines]
            return [engs, Arch(msg.architecture), msg.cpus, msg.memory, msg.hostname, Platform(msg.platform)]
        elif topic == MessageType.START:
            return [msg.binary_filename, msg.binary, FuzzingEngineInfo.from_pb(msg.engine), ExecMode(msg.exec_mode), FuzzMode(msg.fuzz_mode),
                    CheckMode(msg.check_mode), CoverageMode(msg.coverage_mode), SeedInjectLoc(msg.seed_location),
                    msg.engine_args, [x for x in msg.program_argv], msg.sast_report]
        elif topic == MessageType.DATA:
            return [msg.data]
        else:  # for stop and store_coverage_done nothing to unpack
            return []


class BrokerAgent(NetworkAgent):

    def send_seed(self, id: bytes, typ: SeedType, seed: bytes) -> None:
        """
        Send the given input to the client `id`.

        :param id: raw id of the client
        :param typ: Type of the input
        :param seed: Bytes the of input
        """
        msg = InputSeedMsg()
        msg.type = typ.value
        msg.seed = seed
        self.send_to(id, msg, msg_type=MessageType.INPUT_SEED)

    def send_start(self, id: bytes, name: str, package: PathLike, argv: List[str], exmode: ExecMode, fuzzmode: FuzzMode,
                   ckmode: CheckMode, covmode: CoverageMode, engine: FuzzingEngineInfo, engine_args: str,
                   seed_loc: SeedInjectLoc, sast_report: bytes = None) -> None:
        """
        Send a START message to a fuzzing agent with all the parameters it is meant to run with.

        :param id: raw id of the client
        :param name: name of the executable file or binary package
        :param package: filepath of :py:obj:`BinaryPackage` or program executable to send
        :param argv: argumnets to be provided on command line
        :param exmode: execution mode
        :param fuzzmode: fuzzing mode
        :param ckmode: checking mode
        :param covmode: coverage metric to use
        :param engine: descriptor of the fuzzing engine
        :param engine_args: engine's additional arguments or configuration file
        :param seed_loc: location where to provide inputs (stdin or argv)
        :param sast_report: SAST report if applicable
        """
        msg = StartMsg()
        if isinstance(package, str):
            package = Path(package)
        msg.binary_filename = name
        msg.binary = package.read_bytes()
        msg.engine.name = engine.name
        msg.engine.version = engine.version
        msg.exec_mode = exmode.value
        msg.fuzz_mode = fuzzmode.value
        msg.check_mode = ckmode.value
        msg.coverage_mode = covmode.value
        msg.seed_location = seed_loc.value
        msg.engine_args = engine_args
        if sast_report is not None:
            msg.sast_report = sast_report
        for arg in argv:
            msg.program_argv.append(arg)
        self.send_to(id, msg, msg_type=MessageType.START)

    def send_stop(self, id: bytes) -> None:
        """
        Send a stop message to the client.

        :param id: raw id of the client
        """
        msg = StopMsg()
        self.send_to(id, msg, msg_type=MessageType.STOP)

    def register_seed_callback(self, cb: Callable) -> None:
        self.register_callback(MessageType.INPUT_SEED, cb)

    def register_hello_callback(self, cb: Callable) -> None:
        self.register_callback(MessageType.HELLO, cb)

    def register_log_callback(self, cb: Callable) -> None:
        self.register_callback(MessageType.LOG, cb)

    def register_telemetry_callback(self, cb: Callable) -> None:
        self.register_callback(MessageType.TELEMETRY, cb)

    def register_stop_coverage_callback(self, cb: Callable) -> None:
        self.register_callback(MessageType.STOP_COVERAGE_DONE, cb)

    def register_data_callback(self, cb: Callable) -> None:
        self.register_callback(MessageType.DATA, cb)


class ClientAgent(NetworkAgent):
    """
    Subclass of NetworkAgent to connect to PASTIS as a fuzzing
    agent. The class provides helper methods to interact with
    the broker.
    """

    def send_hello(self, engines: List[FuzzingEngineInfo], arch: Arch = None, platform: Platform = None) -> bool:
        """
        Send the hello message to the broker. `engines` parameter is the list of fuzzing engines
        that "we" as client support. E.g: Pastisd is meant to be an interface for all engines
        locally, so it will advertise multiple engines.

        :param engines: list of engines, the client is able to launch
        :param arch: the architecture supported (if None, local one is used)
        :param platform: the platform supported (if None local one used)
        """
        msg = HelloMsg()
        arch = get_local_architecture() if arch is None else arch
        if arch is None:
            logging.error(f"current architecture: {platform.machine()} is not supported")
            return False
        plfm = get_local_platform() if platform is None else platform
        if plfm is None:
            logging.error(f"current platform is not supported")
            return False
        msg.architecture = arch.value
        msg.cpus = psutil.cpu_count()
        msg.memory = psutil.virtual_memory().total
        msg.hostname = socket.gethostname()
        msg.platform = plfm.value
        for eng in engines:
            msg.engines.add(name=eng.name, version=eng.version, pymodule=eng.pymodule)
        self.send(msg, msg_type=MessageType.HELLO)

    def send_log(self, level: LogLevel, message: str) -> None:
        """
        Log message to be sent and printed by the broker. All
        logs received by the broker are logged in a client specific
        logfile.

        :param level: level of the log message
        :param message: message as a string
        """
        self.send(LogMsg(level=level.value, message=message), MessageType.LOG)

    def debug(self, message: str) -> None:
        """
        Send a debug message to the broker

        :param message: message as a string
        """
        self.send_log(LogLevel.DEBUG, message)

    def info(self, message: str) -> None:
        """
        Send an info (level) message to the broker

        :param message: message to send
        """
        self.send_log(LogLevel.INFO, message)

    def warning(self, message: str) -> None:
        """
        Send a warning (level) message to the broker.

        :param message: message to send
        """
        self.send_log(LogLevel.WARNING, message)

    def error(self, message: str) -> None:
        """
        Send an error (level) message to the broker.

        :param message: message to send
        """
        self.send_log(LogLevel.ERROR, message)

    def critical(self, message: str) -> None:
        """
        Send a critical (level) message to the broker.

        :param message: message to send
        """
        self.send_log(LogLevel.CRITICAL, message)

    def send_telemetry(self,
                       state: State = None,
                       exec_per_sec: int = None,
                       total_exec: int = None,
                       cycle: int = None,
                       timeout: int = None,
                       coverage_block: int = None,
                       coverage_edge: int = None,
                       coverage_path: int = None,
                       last_cov_update: int = None) -> None:
        """
        Send a telemetry message to the broker. These data could be used on the
        broker side to plot statistics.

        :param state: current state of the fuzzer
        :param exec_per_sec: number of execution per seconds
        :param total_exec: total number of executions
        :param cycle: number of cycles
        :param timeout: timeout numbers
        :param coverage_block: coverage count in blocks
        :param coverage_edge: coverage count in edges
        :param coverage_path: coverage count in paths
        :param last_cov_update: last coverage update
        """
        msg = TelemetryMsg()
        msg.cpu_usage = psutil.cpu_percent()
        msg.mem_usage = psutil.virtual_memory().percent
        if state:
            msg.state = state.value
        if exec_per_sec:
            msg.exec_per_sec = exec_per_sec
        if total_exec:
            msg.total_exec = total_exec
        if cycle:
            msg.cycle = cycle
        if timeout:
            msg.timeout = timeout
        if coverage_block:
            msg.coverage_block = coverage_block
        if coverage_edge:
            msg.coverage_edge = coverage_edge
        if coverage_path:
            msg.coverage_path = coverage_path
        if last_cov_update:
            msg.last_cov_update = last_cov_update
        self.send(msg, msg_type=MessageType.TELEMETRY)

    def send_stop_coverage_criteria(self) -> None:
        """
        Send a message to the broker indicating, the program has been fully
        covered in accordance to the coverage criteria (metric).
        """
        self.send(StopCoverageCriteria(), MessageType.STOP_COVERAGE_DONE)

    def send_seed(self, typ: SeedType, seed: bytes) -> None:
        """
        Send an input seed to the broker. The ``typ`` indicates
        the type of the seed, namely, input, crash or hang.

        :param typ: type of the input
        :param seed: bytes of the input
        """
        msg = InputSeedMsg()
        msg.type = typ.value
        msg.seed = seed
        self.send(msg, msg_type=MessageType.INPUT_SEED)

    def send_alert_data(self, alert_data: AlertData) -> None:
        """
        Send information related to the coverage or validation of a specific SAST
        alert.

        :param alert_data: alert object
        """
        msg = DataMsg()
        msg.data = alert_data.to_json()
        self.send(msg, msg_type=MessageType.DATA)

    def register_start_callback(self, cb: Callable) -> None:
        """
        Register a callback that will be called when a start message will
        be received. The callback should take 11 parameters.

        :param cb: callback function.
        """
        self.register_callback(MessageType.START, cb)

    def register_stop_callback(self, cb: Callable) -> None:
        """
        Register a callback called when the broker send a STOP
        message. The fuzzing has to stop running and sending data.

        :param cb: callback function
        """
        self.register_callback(MessageType.STOP, cb)

    def register_seed_callback(self, cb: Callable) -> None:
        """
        Register a callback called when an input seed is received from the
        broker. The callback function take 2 parameters seed type and content.

        :param cb: callback function
        """
        self.register_callback(MessageType.INPUT_SEED, cb)

    def register_data_callback(self, cb: Callable) -> None:
        """
        Register callback called when data is received. At the moment
        data are necessarily AlertData messages.

        :param cb: callback function
        """
        self.register_callback(MessageType.DATA, cb)



class FileAgent(ClientAgent):
    """
    Mock agent that will mimick all APIs function of a network agent
    but which will never receive any incoming messages. All messages
    sent are logged to a file
    """

    def __init__(self, level=logging.INFO, log_file: str = None):
        super(FileAgent, self).__init__()
        del self.ctx    # Remove network related attributes
        del self.socket
        self.logger = logging.getLogger('FileAgent')
        self.logger.parent = None  # Remove root handler to make sur it is not printed on output

        # create file handler
        if log_file is not None:
            ch = logging.FileHandler(log_file)
            ch.setLevel(level)
            ch.setFormatter(logging.Formatter('%(asctime)s - [%(name)s] [%(levelname)s]: %(message)s'))
            self.logger.addHandler(ch)

    def bind(self, port: int = 5555, ip: str = "*"):
        raise RuntimeError("FileAgent is not meant to be used as broker")

    def connect(self, remote: str = "localhost", port: int = 5555) -> bool:
        return True  # Do nothing

    def _recv_loop(self):
        while 1:
            if self._stop:
                return
            time.sleep(0.05)

    def send_to(self, id: bytes, msg: Message, msg_type: MessageType = None):
        raise RuntimeError("FileAgent is not meant to be used as broker")

    def send(self, msg: Message, msg_type: MessageType = None):
        if self.mode == AgentMode.BROKER:
            logging.error(f"cannot use sento() as {AgentMode.BROKER.name}")
            return
        if msg_type is None:
            msg_type = self.msg_to_type(msg)

        if isinstance(msg, InputSeedMsg):
            msg = f"{SeedType(msg.type).name}: {msg.seed[:20]}.."
        elif isinstance(msg, HelloMsg):
            msg = f"{msg.hostname}: {Platform(msg.platform)}({Arch(msg.architecture)}) CPU:{msg.cpus} engines:{[x.name for x in msg.engines]}"
        elif isinstance(msg, TelemetryMsg):
            msg = f"{State(msg.state).name} exec/s: {msg.exec_per_sec} total:{msg.total_exec}"
        elif isinstance(msg, LogMsg):
            msg = f"{LogLevel(msg.level).name}: {msg.message}"
        elif isinstance(msg, DataMsg):
            msg = f"Data: {msg.data}"
        elif isinstance(msg, StopCoverageCriteria):
            msg = ""
        else:
            logging.error(f"invalid message type: {type(msg)} as client")
            return

        self.logger.info(f"send {msg_type.name} {msg}")



#!/usr/bin/env python3
# coding: utf-8
"""Installation script for PASTIS."""

from setuptools import setup

with open("README.md") as f:
    lines = f.readlines()
    README = "\n".join(lines[4:7]+lines[51:])


setup(
    name="pastis-framework",
    version="1.0.4",
    description="PASTIS framework for collaborative fuzzing",
    long_description=README,
    long_description_content_type='text/markdown',
    packages=[
        "libpastis",
        "libpastis.proto",
        "pastisbroker",
        "pastisbenchmark",
        "pastisaflpp",
        "aflppbroker",
        "pastishf",
        "hfbroker",
        "pastisdse",
        "pastisttbroker"
    ],
    package_dir={
        # AFL++
        "pastisaflpp": "engines/pastis-aflpp/pastisaflpp",
        "aflppbroker": "engines/pastis-aflpp/broker-addon/aflppbroker",
        # Honggfuzz
        "pastishf": "engines/pastis-honggfuzz/pastishf",
        "hfbroker": "engines/pastis-honggfuzz/broker-addon/hfbroker",
        # Triton
        "pastisdse": "engines/pastis-triton/pastisdse",
        "pastisttbroker": "engines/pastis-triton/broker-addon/pastisttbroker"
    },
    url="https://github.com/quarkslab/pastis",
    project_urls={
        "Documentation": "https://quarkslab.github.io/pastis/",
        "Bug Tracker": "https://github.com/quarkslab/pastis/issues",
        "Source": "https://github.com/quarkslab/pastis"
    },
    setup_requires=[],
    install_requires=[
        "protobuf",
        "pyzmq",
        "psutil",
        "aenum",
        "lief",
        "python-magic",
        "click",
        "coloredlogs",
        "quokka-project",
        "watchdog",
        "pydantic",
        "matplotlib",
        "joblib",
        "rich",
        "tritondse",
    ],
    tests_require=[],
    license="AGPL-3.0",
    author="Quarkslab",
    classifiers=[
        'Topic :: Security',
        'Environment :: Console',
        'Operating System :: OS Independent',
    ],
    test_suite="",
    scripts=[
        'bin/pastis-broker',
        'bin/pastis-benchmark',
        'bin/pastisd',
        'engines/pastis-honggfuzz/bin/pastis-honggfuzz',
        'engines/pastis-triton/bin/pastis-triton',
        'engines/pastis-aflpp/bin/pastis-aflpp'
    ]
)



# Built-in imports
import time
from pathlib import Path
import csv
import json
from typing import List

# Third-party importq
from libpastis.types import SeedType

# Local imports
from pastisbroker.client import PastisClient
from pastisbroker.workspace import Workspace


class StatManager(object):
    """
    Keeps temporal statistics to plot them.
    """
    def __init__(self, workspace: Workspace):
        # Configure CSV writer that will write stats
        names = ['date', 'id', 'exec_per_sec', 'total_exec', 'cycle', 'timeout', 'block', 'edge', 'path', 'last_cov_update']
        self._tel_file = open(workspace.telemetry_file, "w")
        self.writer = csv.DictWriter(self._tel_file, fieldnames=names)
        self.writer.writeheader()

    def update_seed_stat(self, client: PastisClient, typ: SeedType) -> None:
        t = time.localtime()
        if typ == SeedType.INPUT:
            client.input_submitted_count += 1
            client._timeline_seeds.append((t, client.input_submitted_count, typ))
        elif typ == SeedType.CRASH:
            client.crash_submitted_count += 1
            client._timeline_seeds.append((t, client.crash_submitted_count, typ))
        elif typ == SeedType.HANG:
            client.timeout_submitted_count += 1
            client._timeline_seeds.append((t, client.timeout_submitted_count, typ))
        else:
            assert False

        client.seed_first += 1

    def set_exec_per_sec(self, client: PastisClient, exec_per_sec: int = None):
        if exec_per_sec is not None:
            client.exec_per_sec = exec_per_sec  # instantaneous value does not keep history

    def set_total_exec(self, client: PastisClient, total_exec: int = None):
        if total_exec is not None:
            client.total_exec = total_exec  # instantaneous value does not keep history

    def set_cycle(self, client: PastisClient, cycle: int = None):
        if cycle is not None:
            client.cycle = cycle  # instantaneous value does not keep history

    def set_timeout(self, client: PastisClient, timeout: int = None):
        if timeout is not None:
            client.timeout = timeout  # instantaneous value does not keep history

    def set_coverage_block(self, client: PastisClient, coverage: int = None):
        if coverage is not None:
            client.coverage_block = coverage

    def set_coverage_edge(self, client: PastisClient, coverage: int = None):
        if coverage is not None:
            client.coverage_edge = coverage

    def set_coverage_path(self, client: PastisClient, coverage: int = None):
        if coverage is not None:
            client.coverage_path = coverage

    @staticmethod
    def set_last_coverage_update(client: PastisClient, last_up: int = None):
        if last_up is not None:
            client.last_cov_update = last_up  # instantaneous value does not keep history

    def update_telemetry_client(self, client: PastisClient):
        self.writer.writerow({
            'date': time.time(),
            'id': client.strid,
            'exec_per_sec': client.exec_per_sec,
            'total_exec': client.total_exec,
            'cycle': client.cycle,
            'timeout': client.timeout,
            'block': client.coverage_block,
            'edge': client.coverage_edge,
            'path': client.coverage_path,
            'last_cov_update': client.last_cov_update
        })


    def post_execution(self, clients: List[PastisClient], workspace: Workspace) -> None:
        """
        Called at the end of the execution. Export
        :return: None
        """
        self._tel_file.flush()  # Flush the csv if it has not been

        with open(workspace.clients_stat_file, "w") as f:
            json.dump([cli.to_dict() for cli in clients if cli.is_running()], f, indent=2)



# builtin imports
import platform
from typing import Optional

# Local imports
from .types import Arch, Platform


def get_local_architecture() -> Optional[Arch]:
    mapping = {"i386": Arch.X86, "x86_64": Arch.X86_64, "armv7l": Arch.ARMV7, "aarch64": Arch.AARCH64}
    # FIXME: Make sure platform.machine() returns this string for architectures
    return mapping.get(platform.machine())


def get_local_platform() -> Optional[Platform]:
    mapping = {"Linux": Platform.LINUX, "Windows": Platform.WINDOWS, "MacOS": Platform.MACOS, "iOS": Platform.IOS}
    # FIXME: Make sure platform.system() returns this string for other platforms
    return mapping.get(platform.system())



import subprocess
from io import BytesIO
from typing import Optional, List, Tuple
import re
import logging


EXAMPLES = '''
REGEX_1:
==373876==ERROR: AddressSanitizer: stack-buffer-overflow on address 0x7ffecfb907a4 at pc 0x00000043e9e7 bp 0x7ffecfb90660 sp 0x7ffecfb8fdf8

REGEX_2:
==372317==AddressSanitizer: WARNING: unexpected format specifier in printf interceptor: %� (reported once per process)
==372317==AddressSanitizer CHECK failed: /build/llvm-toolchain-9-NoMHhU/llvm-toolchain-9-9.0.1/compiler-rt/lib/asan/../sanitizer_common/sani
'''


class Replay(object):

    INTRINSIC_REGEX = rb".*REACHED ID (\d+)"
    ASAN_REGEX_1 = rb"^==\d+==ERROR: AddressSanitizer: (\S+) (.*)"
    ASAN_REGEX_2 = rb"^==\d+==AddressSanitizer:? ([^:]+): (.*)"
    HF_FETCH = b"HonggfuzzFetchData()"

    def __init__(self):
        self._process = None
        self._alert_covered = []
        self._alert_crash = None
        self._is_hang = False
        self._asan_line = ""
        self._asan_bugtype = ""
        self._is_hf_fetch = False

        # For debugging
        self.stdout, self.stderr = None, None

    @staticmethod
    def run(binary_path: str, args: List[str] = [], stdin_file=None, timeout=None, cwd=None):
        replay = Replay()
        replay._process = subprocess.Popen([binary_path]+args, stdin=open(stdin_file, 'rb'), stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
        try:
            replay.stdout, replay.stderr = replay._process.communicate(timeout=timeout)
            found = replay.__parse_output(replay.stdout)  # In case intrinsic are on output
            found |= replay.__parse_output(replay.stderr)

            if not found and replay.has_crashed():  # Crash that we were not able to link to an ASAN error
                if replay._alert_covered: # Thus take the latest alert and consider it is the origin
                    replay._alert_crash = replay._alert_covered[-1]
        except subprocess.TimeoutExpired:
            replay._is_hang = True

        return replay

    def is_asan_without_crash(self) -> bool:
        """ Return True if an ASAN WARNING was shown without errors """
        return self._asan_bugtype and not self.has_crashed()

    def is_hf_iter_crash(self) -> bool:
        """
        Check we did not hit a call to HF_ITER that would lead
        to such unexpected return code -1
        """
        return self._is_hf_fetch and self.has_crashed()

    @property
    def returncode(self) -> int:
        return self._process.returncode

    @property
    def alert_covered(self) -> List[int]:
        """ Alert covered """
        return self._alert_covered

    def has_hanged(self) -> bool:
        """ Return true if the target hanged during its replay """
        return self._is_hang

    def has_crashed(self) -> bool:
        """ Return whether the execution has crashed or not """
        if self._process.returncode:
            return self._process.returncode != 0
        else:
            return False

    @property
    def crashing_id(self) -> Optional[int]:
        """ Return the alert identifier that made the program to crash (last one seen) """
        return self._alert_crash

    def asan_info(self) -> Tuple[str, str]:
        """ In case of crash return ASAN info gather by parsing """
        return self._asan_bugtype, self._asan_line


    def __parse_output(self, raw_output: bytes) -> bool:
        """ Return True if a vuln was matched """
        matched_vuln = False
        for line in BytesIO(raw_output).readlines():
            # Check if its a line of intrinsic output
            m = re.match(self.INTRINSIC_REGEX, line)
            if m:
                id = int(m.groups()[0])
                self._alert_covered.append(id)

            # Check if its a line of ASAN
            m1 = re.match(self.ASAN_REGEX_1, line)
            m2 = re.match(self.ASAN_REGEX_2, line)
            if m1 or m2:
                if matched_vuln:
                    logging.warning(f"already matched ASAN with {self._asan_bugtype} now {m1.groups() if m1 else m2.groups()}")
                    continue
                matched_vuln = True
                if self._alert_covered: # Try getting last alert (and consider it to be the origin)
                    self._alert_crash = self._alert_covered[-1]
                # Else cannot link the crash to an ID

                # Extract end of line and parameter
                topic, details = m1.groups() if m1 else m2.groups()
                self._asan_bugtype = topic.decode(errors="replace")
                self._asan_line = details.decode(errors="replace")

            if self.HF_FETCH in line:  # In case it is an issue with a HF_ITER read
                self._is_hf_fetch = True

        return matched_vuln



import os
from pathlib import Path
from typing import Optional
import subprocess
import logging

from .driver import AFLPPDriver
from .replay import Replay
from .aflpp import AFLPPProcess, AFLPPNotFound
from .workspace import Workspace

__version__ = "1.0.0"


# AFL++ env variables
AFLPP_ENV_VAR = "AFLPP_WS"

def spawn_online_aflpp(workspace: Optional[Path], port: int = 5555):
    env = os.environ
    env[AFLPP_ENV_VAR] = str(workspace.absolute())
    logging.info(f"aflpp workspace: {str(workspace.absolute())}")
    cmd_line_afl = ["pastis-aflpp", "online", "-p", f"{port}"]
    logging.info(f"run: {' '.join(cmd_line_afl)}")
    return subprocess.Popen(cmd_line_afl, env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE)


def check_scaling_frequency() -> bool:
    data = Path("/sys/devices/system/cpu/cpu0/cpufreq/scaling_governor").read_text()
    return data == "performance\n"



from typing import Tuple
import subprocess

from .pastisdse import PastisDSE

# Expose triton version
import tritondse

__version__ = "1.0.0"

TRITON_VERSION = tritondse.TRITON_VERSION


def spawn_online_triton(port: int = 5555, probe: Tuple[str] = ()):
    tt = ["pastis-triton", "online", "-p", f"{port}"]
    if len(probe) > 0:
        tt += ["--probe", f"{probe}"]
    subprocess.Popen(tt, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
